#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/queue.h>
#include "ptprobes.h"

static const char *prog;
static void usage()
{
	fprintf(stderr, "Usage: %s pid/addr ...\n", prog);
	exit(1);
}

#define container_of(member_addr, container_type, member_name) ({ \
	container_type *c = (container_type*) 0;		\
	unsigned long offset = (unsigned long)&c->member_name;	\
	(container_type*) (((char*) member_addr) - offset);	\
})

struct process_info {
	struct ptp_process pc;
	int running;
	TAILQ_ENTRY(process_info) list;
};
struct probe_info {
	struct ptp_probe pb;
	int nhits;
	TAILQ_ENTRY(probe_info) list;
};
static TAILQ_HEAD(process_head, process_info) process_q;
static TAILQ_HEAD(probe_head, probe_info) probe_q;

#define ckerr(err, func) if (err != 0) { fprintf(stderr, "%s returned %d\n", func, err); exit(2); } else fprintf(stderr, "%s succeeded\n", func);

#define trace_entry() fprintf(stderr, "Entering %s\n", __FUNCTION__)

void probe_handler(struct ptp_probe *p)
{
	struct probe_info *pb = container_of(p, struct probe_info, pb);
	pb->nhits++;
}

void attach_handler(struct ptp_process *pc)
{
	int err;
	struct probe_info *pb;
	fprintf(stderr, "Registering probes for process %d\n", pc->pid);
	for (pb = probe_q.tqh_first; pb; pb = pb->list.tqe_next) {
		if (pb->pb.pid == pc->pid) {
			err = ptp_register_probe(&pb->pb);
			ckerr(err, "ptp_register_probe");
		}
	}
}

void exit_handler(struct ptp_process *p)
{
	struct process_info *pc = container_of(p, struct process_info, pc);
	struct probe_info *pb;

	fprintf(stderr, "Process %d exited.\n", p->pid);
	pc->running = 0;
	for (pb = probe_q.tqh_first; pb; pb = pb->list.tqe_next) {
		if (pb->pb.pid == p->pid) {
			fprintf(stderr, "Probe %d/%p counted %d hits\n",
				pb->pb.pid, pb->pb.addr, pb->nhits);
		}
	}
	
	// Exit if all of our processes have exited.
	for (pc = process_q.tqh_first; pc; pc = pc->list.tqe_next) {
		if (pc->running)
			return;
	}
	exit(0);
}

static void prepare_probe(pid_t pid, ptp_addr_t addr)
{
	struct process_info *pc;
	struct probe_info *pb;

	fprintf(stderr, "Preparing probe %d/%p\n", pid, addr);
	pb = (struct probe_info*) malloc(sizeof(struct probe_info));
	assert(pb != NULL);
	pb->nhits = 0;
	pb->pb.pid = pid;
	pb->pb.addr = addr;
	pb->pb.handler = probe_handler;
	TAILQ_INSERT_TAIL(&probe_q, pb, list);

	// Add this process to the list if it's not already there.
	for (pc = process_q.tqh_first; pc; pc = pc->list.tqe_next) {
		if (pc->pc.pid == pid)
			return;
	}
	pc = (struct process_info*) malloc(sizeof(struct process_info));
	assert(pc != NULL);
	pc->pc.pid = pid;
	pc->pc.attach_handler = attach_handler;
	pc->pc.exit_handler = exit_handler;
	pc->running = -1;
	TAILQ_INSERT_TAIL(&process_q, pc, list);
}

main(int argc, char **argv)
{
	int i, err;
	char *p, *a, *endptr, *slash;
	pid_t pid;
	unsigned long addr;
	struct process_info *pc;

	prog = argv[0];
	if (argc < 2)
		usage();

	TAILQ_INIT(&process_q);
	TAILQ_INIT(&probe_q);

	for (i = 1; i < argc; i++) {
		p = argv[i];
		slash = strchr(p, '/');
		if (!slash)
			usage();
		*slash = '\0';
		a = slash+1;
		if ((pid = atoi(p)) <= 1)
			usage();
		addr = strtoul(a, &endptr, 0);
		if (endptr == a || *endptr != '\0')
			usage();
		prepare_probe(pid, (ptp_addr_t) addr);
	}

	err = ptp_init();
	ckerr(err, "ptp_init");

	for (pc = process_q.tqh_first; pc; pc = pc->list.tqe_next) {
		pc->running = 1;
		err = ptp_attach_process(&pc->pc);
		ckerr(err, "ptp_attach_process");
	}

	ptp_run();
	/*NOTREACHED*/
}
