/*
 * ptrace-based user-space probes
 * Copyright (C) International Business Machines Corp., 2006
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  Project Website:  http://sourceware.org/systemtap
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <malloc.h>
#include <assert.h>
#include <sys/ptrace.h>
#include <sys/queue.h>

#include "ptprobes_impl.h"

/* Queue of events captured by our SIGCHLD handler */
static TAILQ_HEAD(sigchld_head, sigchld) sigchld_q;
static long nsigchld = 0;

#define HASH_BITS 6
#define HASH_SIZE (1 << HASH_BITS)
/* Hash of probed processes */
LIST_HEAD(process_bucket, ptp_process);
static struct process_bucket process_hash[HASH_SIZE];

/* TODO: Consider different hash table sizes for processes and probes. */
/* TODO: Consider other hash algorithms. Remember that we're LGPL, not GPL.*/
/* TODO: GPL or LGPL? */
#define ADDR_HASH(addr) ( (((unsigned long)(addr)) >> 2) % HASH_SIZE )
#define PID_HASH(pid) ((pid) % HASH_SIZE)

/* TODO: Refuse to probe instructions such as int3. */
static opcode_t replace_opcode(pid_t pid, ptp_addr_t addr, opcode_t opcode);

static struct ptp_process *find_process(pid_t pid)
{
	struct process_bucket *head = &process_hash[PID_HASH(pid)];
	struct ptp_process *p;

	for (p = head->lh_first; p; p = p->list.le_next) {
		if (p->pid == pid)
			return p;
	}
	return NULL;
}

static struct ptp_aggr_probe *find_aggr(struct ptp_process *process,
	ptp_addr_t addr)
{
	struct ptp_aggr_probe_bucket *head;
	struct ptp_aggr_probe *p;
	head = &process->probe_hash[ADDR_HASH(addr)];
	for (p = head->lh_first; p; p = p->list.le_next) {
		if (p->addr == addr)
			return p;
	}
	return NULL;
}

static int initialize_aggr(struct ptp_aggr_probe *aggr, struct ptp_probe *p)
{
	aggr->pid = p->pid;
	aggr->addr = p->addr;
	aggr->opcode = replace_opcode(aggr->pid, aggr->addr,
		BREAKPOINT_INSTRUCTION);
	if (errno)
		return errno;
	TAILQ_INIT(&aggr->probes);
	return 0;
}

/*
 * If an aggregate probe matching p already exists, just add p to it.
 * Otherwise, create the aggregate probe and install the breakpoint first.
 */
static int install_probe(struct ptp_probe *p, struct ptp_process *process)
{
	struct ptp_aggr_probe *aggr = find_aggr(process, p->addr);
	if (!aggr) {
		int err;
		struct ptp_aggr_probe_bucket *head;
		aggr = (struct ptp_aggr_probe*)
			malloc(sizeof(struct ptp_aggr_probe));
		if (!aggr)
			return ENOMEM;
		err = initialize_aggr(aggr, p);
		if (err != 0) {
			free(aggr);
			return err;
		}
		head = &process->probe_hash[ADDR_HASH(p->addr)];
		LIST_INSERT_HEAD(head, aggr, list);
	}
	TAILQ_INSERT_TAIL(&aggr->probes, p, list);
	p->aggr = aggr;
	return 0;
}

static void block_signals(sigset_t *oldset)
{
	sigset_t all_sigs;

	(void) sigfillset(&all_sigs);
	sigprocmask(SIG_BLOCK, &all_sigs, oldset);
}

static void restore_signals(sigset_t *oldset)
{
	sigprocmask(SIG_SETMASK, oldset, NULL);
}

#if 0
/* This slows things down about 30% and doesn't seem to help. */
long safe_ptrace(enum __ptrace_request request, pid_t pid, void *addr, void
	*data)
{
	long ret;
	sigset_t oldset;
	block_signals(&oldset);
	ret = ptrace(request, pid, addr, data);
	restore_signals(&oldset);
	return ret;
}
#define _ptrace safe_ptrace
#endif
#define _ptrace ptrace

/*
 * Optimization: Add a version of replace_opcode() that omits the
 * PTRACE_PEEKTEXT call.  This can be used if we "know" what was there
 * originally.  This is the case every time but the first time.  However,
 * we'd have to keep track of any probepoints that fall within sizeof(long)
 * of each other.
 */

/*
 * Returns the original opcode.  Caller should check errno.
 */
static opcode_t replace_opcode(pid_t pid, ptp_addr_t addr, opcode_t opcode) {
	long inst;
	opcode_t *pinst = (opcode_t*) &inst;
	opcode_t prev_opcode;

	errno = 0;
	inst = _ptrace(PTRACE_PEEKTEXT, pid, (void*)addr, NULL);
	if (errno)
		return (opcode_t) errno;
	prev_opcode = pinst[0];
	pinst[0] = opcode;
	if (_ptrace(PTRACE_POKETEXT, pid, (void*)addr, (void*)inst) != 0)
		return (opcode_t) errno;
	return prev_opcode;
}

static void report_sigchld(const struct sigchld *s)
{
	fprintf(stderr,
		"SIGCHLD #%ld: pid=%d, signo=%d, realsig=%d, reason=%d\n",
		s->id, s->pid, s->signo, s->realsig, s->reason);
}

/*
 * SIGCHLD handler.  Populates a sigchld object and queues it.
 * All signals are blocked while this handler runs.
 */
static void sigchld(int sig, siginfo_t *si, void *data)
{
	struct sigchld *s = (struct sigchld *) malloc(sizeof(struct sigchld));
	assert(s != NULL);
	s->signo = si->si_signo;
	s->reason = si->si_code;
	s->realsig = si->si_status;
	s->pid = si->si_pid;
	s->id = ++nsigchld;
	TAILQ_INSERT_TAIL(&sigchld_q, s, list);
}

int ptp_init(void)
{
	struct sigaction act;
	int i;
	int err;

	TAILQ_INIT(&sigchld_q);

	for (i = 0; i < HASH_SIZE; i++) {
		LIST_INIT(&process_hash[i]);
	}

	/* Set up SIGCHLD handler. */
	(void) memset(&act, 0, sizeof(act));
	act.sa_sigaction = sigchld;
	(void) sigfillset(&act.sa_mask);	// Block all sigs for hander.
	act.sa_flags = SA_SIGINFO;
	err = sigaction(SIGCHLD, &act, NULL);
	if (err != 0)
		return errno;
	return 0;
}

int ptp_attach_process(struct ptp_process *p)
{
	int err, i;
	if (!p)
		return EINVAL;
	p->state = ptpc_attached;
	p->active_probe = NULL;
	err = _ptrace(PTRACE_ATTACH, p->pid, 0, 0);
	if (err != 0)
		return errno;

	p->probe_hash = (struct ptp_aggr_probe_bucket*)
		malloc(HASH_SIZE * sizeof(struct ptp_aggr_probe_bucket));
	if (!p->probe_hash)
		return ENOMEM;
	for (i = 0; i < HASH_SIZE; i++)
		LIST_INIT(&p->probe_hash[i]);

	LIST_INSERT_HEAD(&process_hash[PID_HASH(p->pid)], p, list);
	return 0;
}

/*
 * p->pid should be stopped when this is called.
 */
int ptp_register_probe(struct ptp_probe *p)
{
	struct ptp_process *process;

	if (!p)
		return EINVAL;
	process = find_process(p->pid);
	if (!process)
		return ESRCH;
	if (process->state != ptpc_probing)
		return ESRCH;

	return install_probe(p, process);
}

/*
 * p->pid should be stopped when this is called.
 */
void ptp_unregister_probe(struct ptp_probe *p)
{
	// TODO
	// Find the probe.  Could chase p->aggr, but not safe if p is bad.
	// Remove the probe from p->aggr.
	// If that was p->aggr's last probe, restore the opcode and
	// 	unlink/free p->aggr.
}

// TODO: Minimize block/restore.  E.g., block only if we first see
// that s is not NULL.
static struct sigchld *pop_sigchld()
{
	struct sigchld *s;
	sigset_t oldset;
	block_signals(&oldset);
	s = sigchld_q.tqh_first;
	if (s)
		TAILQ_REMOVE(&sigchld_q, s, list);
	restore_signals(&oldset);
	return s;
}

/*
 * If this is our breakpoint, run the handler, kick off single-stepping,
 * and return 0.  If it's not our breakpoint, return 1.  Return <0 on error.
 */
static int handle_bkpt(struct ptp_process *process)
{
	int err;
	pid_t pid = process->pid;
	unsigned long regs[FRAME_SIZE];
	unsigned long ip, addr;
	struct ptp_aggr_probe *aggr;
	struct ptp_probe *p;

	if (_ptrace(PTRACE_GETREGS, pid, 0, regs) != 0)
		return -1;
	ip = regs[IPREG];
	addr = IP_TO_ADDR(ip);
	
	aggr = find_aggr(process, (ptp_addr_t) addr);
	if (!aggr)
		return 1;
	for (p = aggr->probes.tqh_first; p; p = p->list.tqe_next) {
		if (p->handler)
			(void) p->handler(p);
	}
	(void) replace_opcode(pid, aggr->addr, aggr->opcode);
	if (errno)
		return -1;
	regs[EIP] = addr;
	err = _ptrace(PTRACE_SETREGS, pid, 0, regs);
	if (err != 0)
		return -1;
	err = _ptrace(PTRACE_SINGLESTEP, pid, 0, 0);
	if (err != 0)
		return -1;
	process->active_probe = aggr;
	return 0;
}

/* Resume execution after single-stepping. */
static int resume_execution(struct ptp_process *process,
	struct ptp_aggr_probe *aggr)
{
	(void) replace_opcode(process->pid, aggr->addr, BREAKPOINT_INSTRUCTION);
	if (errno)
		return -1;
	process->active_probe = NULL;
	return _ptrace(PTRACE_CONT, process->pid, 0, 0);
}

/*
 * A child took a SIGTRAP, which came to us as a SIGCHLD.  Typically, it's
 * associated with hitting a breakpoint or single-stepping.  Deal with it.
 */
static int handle_trap(const struct sigchld *s)
{
	int err;
	struct ptp_aggr_probe *aggr;
	struct ptp_process *process = find_process(s->pid);

	if (!process)
		goto unknown_trap;
	aggr = process->active_probe;
	if (aggr)
		return resume_execution(process, aggr);

	err = handle_bkpt(process);
	if (err == 1)
		goto unknown_trap;
	return err;

unknown_trap:
	report_sigchld(s);
	return _ptrace(PTRACE_CONT, s->pid, 0, (void*)s->realsig);
}

/*
 * This is the main event loop, called by the user after attaching to
 * the selected process(es);
 */
void ptp_run(void)
{
	struct sigchld *s;
	struct ptp_process *p;
	int err;

	for (;;) {
		wait();
		while ((s = pop_sigchld()) != NULL) {
			// report_sigchld(s);
			err = 0;
			switch (s->reason) {
			case CLD_TRAPPED:
				if (s->realsig == SIGTRAP)
					err = handle_trap(s);
				else {
					report_sigchld(s);
					err = _ptrace(PTRACE_CONT, s->pid, 0,
						(void*)s->realsig);
				}
				break;
			case CLD_STOPPED:
				report_sigchld(s);
				p = find_process(s->pid);
				if (p && p->state == ptpc_attached) {
					/* SIGSTOP after PTRACE_ATTACH */
					p->state = ptpc_probing;
					if (p->attach_handler)
						p->attach_handler(p);
					// What's correct here?
					// 0, SIGSTOP, and SIGCONT seem to
					// have the same effect.
					// s->realsig = 0;
				}
				/*FALLTHROUGH*/
			case CLD_CONTINUED:
				err = _ptrace(PTRACE_CONT, s->pid, 0,
					(void*)s->realsig);
				break;
			case CLD_EXITED:
			case CLD_KILLED:
			case CLD_DUMPED:
				report_sigchld(s);
				p = find_process(s->pid);
				if (p && p->state != ptpc_exited) {
					p->state = ptpc_exited;
					if (p->exit_handler)
						p->exit_handler(p);
				}
				// err = _ptrace(PTRACE_DETACH, s->pid, 0, 0);
				break;
			default:
				report_sigchld(s);
				fprintf(stderr, "Unhandled reason: %d\n",
					s->reason);
			}
			if (err != 0) {
				perror("SIGCHLD handling");
				report_sigchld(s);
			}
			free(s);
		}
	}
	/*NOTREACHED*/
}
